@@ -1,16 +1,19 @@
+import asyncio
 import logging
 import re
-from typing import Dict
+from typing import Dict, List
 
 from container_app_conf.formatter.toml import TomlFormatter
 from requests import HTTPError
-from telegram import Update, ParseMode, InlineKeyboardMarkup, InlineKeyboardButton, ReplyKeyboardRemove
-from telegram.ext import CommandHandler, Filters, MessageHandler, Updater, CallbackContext, CallbackQueryHandler
+from telegram import Update, InlineKeyboardMarkup, InlineKeyboardButton, ReplyKeyboardRemove
+from telegram.ext import CommandHandler, filters, MessageHandler, CallbackQueryHandler, \
+    ApplicationBuilder, ContextTypes
 from telegram_click.argument import Argument, Flag
 from telegram_click.decorator import command
 
+from keel_telegram_bot import util
 from keel_telegram_bot.api_client import KeelApiClient
-from keel_telegram_bot.bot.permissions import CONFIG_ADMINS
+from keel_telegram_bot.bot.permissions import CONFIG_ADMINS, CONFIGURED_CHAT_ID
 from keel_telegram_bot.bot.reply_keyboard_handler import ReplyKeyboardHandler
 from keel_telegram_bot.config import Config
 from keel_telegram_bot.stats import *
@@ -31,76 +34,82 @@ class KeelTelegramBot:
         """
         self._config = config
         self._api_client = api_client
+        self._message_map = {}
 
         self._response_handler = ReplyKeyboardHandler()
 
-        self._updater = Updater(token=self._config.TELEGRAM_BOT_TOKEN.value, use_context=True)
-        LOGGER.debug(f"Using bot id '{self._updater.bot.id}' ({self._updater.bot.name})")
-        self._dispatcher = self._updater.dispatcher
+        self._app = ApplicationBuilder().token(self._config.TELEGRAM_BOT_TOKEN.value).build()
 
         handler_groups = {
             0: [CallbackQueryHandler(callback=self._inline_keyboard_click_callback)],
             1: [
                 CommandHandler(COMMAND_START,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._start_callback),
                 CommandHandler(COMMAND_LIST_APPROVALS,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._list_approvals_callback),
                 CommandHandler(COMMAND_APPROVE,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._approve_callback),
                 CommandHandler(COMMAND_REJECT,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._reject_callback),
                 CommandHandler(COMMAND_DELETE,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._delete_callback),
+                CommandHandler(COMMAND_CHATID,
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
+                               callback=self._chatid_callback),
 
                 CommandHandler(COMMAND_HELP,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._help_callback),
                 CommandHandler(COMMAND_CONFIG,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._config_callback),
                 CommandHandler(COMMAND_VERSION,
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._version_callback),
                 CommandHandler(CANCEL_KEYBOARD_COMMAND[1:],
-                               filters=(~ Filters.reply) & (~ Filters.forwarded),
+                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                                callback=self._response_handler.cancel_keyboard_callback),
                 # unknown command handler
                 MessageHandler(
-                    filters=Filters.command & (~ Filters.forwarded),
+                    filters=filters.COMMAND & (~ filters.FORWARDED),
                     callback=self._unknown_command_callback),
                 MessageHandler(
-                    filters=(~ Filters.forwarded),
+                    filters=(~ filters.FORWARDED),
                     callback=self._any_message_callback),
             ]
         }
 
         for group, handlers in handler_groups.items():
             for handler in handlers:
-                self._updater.dispatcher.add_handler(handler, group=group)
+                self._app.add_handler(handler, group=group)
 
     @property
     def bot(self):
-        return self._updater.bot
+        return self._app.bot
 
-    def start(self):
+    async def start(self):
         """
         Starts up the bot.
         """
-        self._updater.start_polling()
+        await self._app.initialize()
+        LOGGER.debug(f"Using bot id '{self._app.bot.id}' ({self._app.bot.name})")
+        await self._app.start()
+        await self._app.updater.start_polling()
 
     def stop(self):
         """
         Shuts down the bot.
         """
-        self._updater.stop()
+        loop = asyncio.get_event_loop()
+        loop.run_until_complete(self._app.shutdown())
 
     @COMMAND_TIME_START.time()
-    def _start_callback(self, update: Update, context: CallbackContext) -> None:
+    async def _start_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
         """
         Welcomes a new user with a greeting message
         :param update: the chat update object
@@ -111,11 +120,11 @@ class KeelTelegramBot:
         user_first_name = update.effective_user.first_name
 
         if not CONFIG_ADMINS.evaluate(update, context):
-            send_message(bot, chat_id, "Sorry, you do not have permissions to use this bot.")
+            await send_message(bot, chat_id, "Sorry, you do not have permissions to use this bot.")
             return
 
-        send_message(bot, chat_id,
-                     f"Welcome {user_first_name},\nthis is your keel-telegram-bot instance, ready to go!")
+        await send_message(bot, chat_id,
+                           f"Welcome {user_first_name},\nthis is your keel-telegram-bot instance, ready to go!")
 
     @COMMAND_TIME_LIST_APPROVALS.time()
     @command(name=COMMAND_LIST_APPROVALS,
@@ -125,9 +134,9 @@ class KeelTelegramBot:
                  Flag(name=["approved", "a"], description="Include approved items"),
                  Flag(name=["rejected", "r"], description="Include rejected items"),
              ],
-             permissions=CONFIG_ADMINS)
-    def _list_approvals_callback(self, update: Update, context: CallbackContext,
-                                 archived: bool, approved: bool, rejected: bool) -> None:
+             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
+    async def _list_approvals_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
+                                       archived: bool, approved: bool, rejected: bool) -> None:
         """
         List pending approvals
         """
@@ -136,6 +145,7 @@ class KeelTelegramBot:
         chat_id = update.effective_chat.id
 
         items = self._api_client.get_approvals()
+        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))
 
         rejected_items = list(filter(lambda x: x[KEY_REJECTED], items))
         archived_items = list(filter(lambda x: x[KEY_ARCHIVED], items))
@@ -174,7 +184,7 @@ class KeelTelegramBot:
         ]))
 
         text = "\n\n".join(lines).strip()
-        send_message(bot, chat_id, text, reply_to=message.message_id, parse_mode=ParseMode.HTML)
+        await send_message(bot, chat_id, text, reply_to=message.message_id, parse_mode="HTML")
 
     @COMMAND_TIME_APPROVE.time()
     @command(name=COMMAND_APPROVE,
@@ -184,32 +194,38 @@ class KeelTelegramBot:
                           example="default/myimage:1.5.5"),
                  Argument(name=["voter", "v"], description="Name of voter", example="john", optional=True),
              ],
-             permissions=CONFIG_ADMINS)
-    def _approve_callback(self, update: Update, context: CallbackContext,
-                          identifier: str, voter: str or None) -> None:
+             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
+    async def _approve_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
+                                identifier: str, voter: str or None) -> None:
         """
         Approve a pending item
         """
+        chat_id = update.effective_chat.id
+
         if voter is None:
             voter = update.effective_user.full_name
 
-        def execute(update: Update, context: CallbackContext, item: dict, data: dict):
+        async def execute(update: Update, context: ContextTypes.DEFAULT_TYPE, item: dict, data: dict):
             bot = context.bot
             message = update.effective_message
             chat_id = update.effective_chat.id
 
             self._api_client.approve(item["id"], item["identifier"], voter)
             text = f"Approved {item['identifier']}"
-            send_message(bot, chat_id, text, reply_to=message.message_id, menu=ReplyKeyboardRemove(selective=True))
+            await send_message(bot, chat_id, text, reply_to=message.message_id,
+                               menu=ReplyKeyboardRemove(selective=True))
 
         items = self._api_client.get_approvals(rejected=False, archived=False)
+        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))
 
+        # compare to the "id" first
         exact_matches = list(filter(lambda x: x["id"] == identifier, items))
         if len(exact_matches) > 0:
-            execute(update, context, exact_matches[0], {})
+            await execute(update, context, exact_matches[0], {})
             return
 
-        self._response_handler.await_user_selection(
+        # then fuzzy match to "identifier"
+        await self._response_handler.await_user_selection(
             update, context, identifier, choices=items, key=lambda x: x["identifier"],
             callback=execute,
         )
@@ -222,32 +238,39 @@ class KeelTelegramBot:
                           example="default/myimage:1.5.5"),
                  Argument(name=["voter", "v"], description="Name of voter", example="john", optional=True),
              ],
-             permissions=CONFIG_ADMINS)
-    def _reject_callback(self, update: Update, context: CallbackContext,
-                         identifier: str, voter: str or None) -> None:
+             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
+    async def _reject_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
+                               identifier: str, voter: str or None) -> None:
         """
         Reject a pending item
         """
+        chat_id = update.effective_chat.id
         if voter is None:
             voter = update.effective_user.full_name
+        if not voter:
+            voter = update.effective_user.name
 
-        def execute(update: Update, context: CallbackContext, item: dict, data: dict):
+        async def execute(update: Update, context: ContextTypes.DEFAULT_TYPE, item: dict, data: dict):
             bot = context.bot
             message = update.effective_message
             chat_id = update.effective_chat.id
 
             self._api_client.reject(item["id"], item["identifier"], voter)
             text = f"Rejected {item['identifier']}"
-            send_message(bot, chat_id, text, reply_to=message.message_id, menu=ReplyKeyboardRemove(selective=True))
+            await send_message(bot, chat_id, text, reply_to=message.message_id,
+                               menu=ReplyKeyboardRemove(selective=True))
 
         items = self._api_client.get_approvals(rejected=False, archived=False)
+        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))
 
+        # compare to the "id" first
         exact_matches = list(filter(lambda x: x["id"] == identifier, items))
         if len(exact_matches) > 0:
-            execute(update, context, exact_matches[0], {})
+            await execute(update, context, exact_matches[0], {})
             return
 
-        self._response_handler.await_user_selection(
+        # then fuzzy match to "identifier"
+        await self._response_handler.await_user_selection(
             update, context, identifier, choices=items, key=lambda x: x["identifier"],
             callback=execute,
         )
@@ -260,37 +283,42 @@ class KeelTelegramBot:
                           example="default/myimage:1.5.5"),
                  Argument(name=["voter", "v"], description="Name of voter", example="john", optional=True),
              ],
-             permissions=CONFIG_ADMINS)
-    def _delete_callback(self, update: Update, context: CallbackContext,
-                         identifier: str, voter: str or None) -> None:
+             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
+    async def _delete_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
+                               identifier: str, voter: str or None) -> None:
         """
         Delete an archived item
         """
+        chat_id = update.effective_chat.id
         if voter is None:
             voter = update.effective_user.full_name
 
-        def execute(update: Update, context: CallbackContext, item: dict, data: dict):
+        async def execute(update: Update, context: ContextTypes.DEFAULT_TYPE, item: dict, data: dict):
             bot = context.bot
             message = update.effective_message
             chat_id = update.effective_chat.id
 
             self._api_client.delete(item["id"], item["identifier"], voter)
             text = f"Deleted {item['identifier']}"
-            send_message(bot, chat_id, text, reply_to=message.message_id, menu=ReplyKeyboardRemove(selective=True))
+            await send_message(bot, chat_id, text, reply_to=message.message_id,
+                               menu=ReplyKeyboardRemove(selective=True))
 
         items = self._api_client.get_approvals()
+        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))
 
+        # compare to the "id" first
         exact_matches = list(filter(lambda x: x["id"] == identifier, items))
         if len(exact_matches) > 0:
-            execute(update, context, exact_matches[0], {})
+            await execute(update, context, exact_matches[0], {})
             return
 
-        self._response_handler.await_user_selection(
+        # then fuzzy match to "identifier"
+        await self._response_handler.await_user_selection(
             update, context, identifier, choices=items, key=lambda x: x["identifier"],
             callback=execute,
         )
 
-    def on_notification(self, data: dict):
+    async def on_notification(self, data: dict):
         """
         Handles incoming notifications (via Webhook)
         :param data: notification data
@@ -311,76 +339,99 @@ class KeelTelegramBot:
         ])
 
         for chat_id in self._config.TELEGRAM_CHAT_IDS.value:
-            send_message(
+
+            if self._is_filtered_for(chat_id, identifier):
+                continue
+
+            await send_message(
                 self.bot, chat_id,
-                text, parse_mode=ParseMode.HTML,
+                text, parse_mode="HTML",
                 menu=None
             )
 
-    def on_new_pending_approval(self, item: dict):
+    async def on_new_pending_approval(self, item: dict):
         """
         Handles new pending approvals by sending a message
         including an inline keyboard to all configured chat ids
         :param item: new pending approval
         """
+        identifier = item["identifier"]
         text = approval_to_str(item)
-
-        keyboard_items = {
-            "Approve": BUTTON_DATA_APPROVE,
-            "Reject": BUTTON_DATA_REJECT
-        }
-        keyboard = self._build_inline_keyboard(keyboard_items)
+        menu = self.create_approval_notification_menu(item)
 
         for chat_id in self._config.TELEGRAM_CHAT_IDS.value:
-            send_message(
-                self.bot, chat_id,
-                text, parse_mode=ParseMode.HTML,
-                menu=keyboard
-            )
+
+            if self._is_filtered_for(chat_id, identifier):
+                LOGGER.debug(f"Skipping new pending approval for chat '{chat_id}' due to filters")
+                continue
+
+            LOGGER.debug(f"Sending pending approval message to '{chat_id}'")
+            try:
+                response = await send_message(
+                    self.bot, chat_id,
+                    text, parse_mode="HTML",
+                    menu=menu
+                )
+                self._register_message(response.chat_id, response.message_id, item["id"], item["identifier"])
+            except Exception as ex:
+                LOGGER.exception(ex)
 
     @command(
         name=COMMAND_CONFIG,
         description="Print bot config.",
-        permissions=CONFIG_ADMINS,
+        permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS,
     )
-    def _config_callback(self, update: Update, context: CallbackContext):
+    async def _config_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
         bot = context.bot
         message = update.effective_message
         chat_id = update.effective_chat.id
         text = self._config.print(TomlFormatter())
-        send_message(bot, chat_id, text, reply_to=message.message_id)
+        await send_message(bot, chat_id, text, reply_to=message.message_id)
 
     @command(
         name=COMMAND_HELP,
         description="List commands supported by this bot.",
         permissions=CONFIG_ADMINS,
     )
-    def _help_callback(self, update: Update, context: CallbackContext):
+    async def _help_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
         bot = context.bot
         message = update.effective_message
         chat_id = update.effective_chat.id
 
         from telegram_click import generate_command_list
-        text = generate_command_list(update, context)
-        send_message(bot, chat_id, text,
-                     parse_mode=ParseMode.MARKDOWN,
-                     reply_to=message.message_id)
+        text = await generate_command_list(update, context)
+        await send_message(bot, chat_id, text,
+                           parse_mode="MARKDOWN",
+                           reply_to=message.message_id)
+
+    @command(
+        name=COMMAND_CHATID,
+        description="Print chat ID.",
+        permissions=CONFIG_ADMINS,
+    )
+    async def _chatid_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
+        bot = context.bot
+        message = update.effective_message
+        chat_id = update.effective_chat.id
+
+        text = update.message.chat_id
+        await send_message(bot, chat_id, str(text), reply_to=message.message_id)
 
     @command(
         name=COMMAND_VERSION,
         description="Print bot version.",
         permissions=CONFIG_ADMINS,
     )
-    def _version_callback(self, update: Update, context: CallbackContext):
+    async def _version_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
         bot = context.bot
         message = update.effective_message
         chat_id = update.effective_chat.id
 
         from keel_telegram_bot import __version__
         text = __version__
-        send_message(bot, chat_id, text, reply_to=message.message_id)
+        await send_message(bot, chat_id, text, reply_to=message.message_id)
 
-    def _unknown_command_callback(self, update: Update, context: CallbackContext) -> None:
+    async def _unknown_command_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
         """
         Handles unknown commands send by a user
         :param update: the chat update object
@@ -393,18 +444,18 @@ class KeelTelegramBot:
 
         user_is_admin = username in self._config.TELEGRAM_ADMIN_USERNAMES.value
         if user_is_admin:
-            self._help_callback(update, context)
+            await self._help_callback(update, context)
             return
 
-    def _any_message_callback(self, update: Update, context: CallbackContext) -> None:
+    async def _any_message_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
         """
         Used to respond to response keyboard entry selections
         :param update: the chat update object
         :param context: telegram context
         """
-        self._response_handler.on_message(update, context)
+        await self._response_handler.on_message(update, context)
 
-    def _inline_keyboard_click_callback(self, update: Update, context: CallbackContext):
+    async def _inline_keyboard_click_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
         """
         Handles inline keyboard button click callbacks
         :param update:
@@ -430,26 +481,23 @@ class KeelTelegramBot:
             if data == BUTTON_DATA_APPROVE:
                 self._api_client.approve(approval_id, approval_identifier, from_user.full_name)
                 answer_text = f"Approved '{approval_identifier}'"
-                keyboard = self._build_inline_keyboard({"Approved": BUTTON_DATA_NOTHING})
                 KEEL_APPROVAL_ACTION_COUNTER.labels(action="approve", identifier=approval_identifier).inc()
             elif data == BUTTON_DATA_REJECT:
                 self._api_client.reject(approval_id, approval_identifier, from_user.full_name)
                 answer_text = f"Rejected '{approval_identifier}'"
-                keyboard = self._build_inline_keyboard({"Rejected": BUTTON_DATA_NOTHING})
                 KEEL_APPROVAL_ACTION_COUNTER.labels(action="reject", identifier=approval_identifier).inc()
             else:
-                bot.answer_callback_query(query_id, text="Unknown button")
+                await bot.answer_callback_query(query_id, text="Unknown button")
                 return
 
-            # remove buttons
-            query.edit_message_reply_markup(reply_markup=keyboard)
-            context.bot.answer_callback_query(query_id, text=answer_text)
+            await context.bot.answer_callback_query(query_id, text=answer_text)
+            await self.update_messages()
         except HTTPError as e:
             LOGGER.error(e)
-            bot.answer_callback_query(query_id, text=f"{e.response.content.decode('utf-8')}")
+            await bot.answer_callback_query(query_id, text=f"{e.response.content.decode('utf-8')}")
         except Exception as e:
             LOGGER.error(e)
-            bot.answer_callback_query(query_id, text=f"Unknwon error")
+            await bot.answer_callback_query(query_id, text=f"Unknwon error")
 
     @staticmethod
     def _build_inline_keyboard(items: Dict[str, str]) -> InlineKeyboardMarkup:
@@ -460,3 +508,78 @@ class KeelTelegramBot:
         """
         keyboard = list(map(lambda x: InlineKeyboardButton(x[0], callback_data=x[1]), items.items()))
         return InlineKeyboardMarkup.from_column(keyboard)
+
+    def create_approval_notification_menu(self, item: dict) -> InlineKeyboardMarkup:
+        keyboard_items = {}
+        if item["archived"]:
+            keyboard_items["Approved"] = BUTTON_DATA_NOTHING
+        elif item["rejected"]:
+            keyboard_items["Rejected"] = BUTTON_DATA_NOTHING
+        else:
+            if item["votesRequired"] > item["votesReceived"]:
+                keyboard_items["Approve"] = BUTTON_DATA_APPROVE
+                keyboard_items["Reject"] = BUTTON_DATA_REJECT
+
+        return self._build_inline_keyboard(keyboard_items)
+
+    def _register_message(self, chat_id: int, message_id: int, approval_id: str, approval_identifier: str):
+        """
+        Registers a telegram message, that corresponds with an approval notification.
+        This is used to update this message. This is possible for approx. 48 hours, after
+        which telegram prohibits modifications of the original message.
+        :param chat_id: chat id
+        :param message_id: message id
+        :param approval_id: approval id
+        :param approval_identifier: approval identifier
+        """
+        key = f"{approval_id}_{approval_identifier}"
+        self._message_map.setdefault(key, {}).setdefault(chat_id, set()).add(message_id)
+
+    async def update_messages(self):
+        """
+        Fetch approvals and update existing approval messages accordingly
+        """
+        approvals = self._api_client.get_approvals()
+
+        for approval in approvals:
+            approval_id = approval["id"]
+            approval_identifier = approval["identifier"]
+            key = f"{approval_id}_{approval_identifier}"
+
+            chats = self._message_map.get(key, {})
+            failed_messages = set()
+            for chat_id, message_ids in chats.items():
+
+                if self._is_filtered_for(chat_id, approval_identifier):
+                    continue
+
+                for message_id in message_ids:
+                    try:
+                        approval_str = approval_to_str(approval)
+                        menu = self.create_approval_notification_menu(approval)
+                        await self.bot.edit_message_text(
+                            approval_str,
+                            chat_id=chat_id,
+                            message_id=message_id,
+                            parse_mode="HTML",
+                            reply_markup=menu
+                        )
+                    except Exception as ex:
+                        failed_messages.add(message_id)
+                        LOGGER.exception(ex)
+
+                for failure in failed_messages:
+                    message_ids.remove(failure)
+
+    def _is_filtered_for(self, chat_id: str or int, identifier: str) -> bool:
+        chat_ids: List[str] = self._config.TELEGRAM_CHAT_IDS.value
+        filter_config: List[Dict] = self._config.TELEGRAM_FILTERS.value
+
+        chat_unknown = str(chat_id) not in chat_ids
+        filter_doesnt_match = util._is_filtered_for(filter_config, str(chat_id), identifier)
+
+        result = chat_unknown or filter_doesnt_match
+        LOGGER.debug(
+            f"Filtered identifier '{identifier}' for chat {chat_id} because: chat_unknown: {chat_unknown}, filter_doesnt_match: {filter_doesnt_match}, result: {result}")
+
+        return result
