@@ -7,6 +7,7 @@ COMMAND_LIST_APPROVALS = ["list", "l"]
 COMMAND_APPROVE = ["approve", "a"]
 COMMAND_REJECT = ["reject", "r"]
 COMMAND_DELETE = ["delete", "d"]
+COMMAND_CHATID = "chatid"
 
 COMMAND_STATS = 'stats'
 
@@ -21,6 +22,10 @@ BUTTON_DATA_NOTHING = "_"
 BUTTON_DATA_APPROVE = "a"
 BUTTON_DATA_REJECT = "r"
 
+# webserver
+
+ENDPOINT_WEBHOOK = "/"
+
 # api keys
 KEY_ARCHIVED = "archived"
 KEY_REJECTED = "rejected"
