@@ -1,11 +1,13 @@
 import functools
 import logging
 import operator
+import re
 from datetime import datetime, timezone, timedelta
-from typing import List, Any, Tuple
+from typing import List, Any, Tuple, Dict
 
 import iso8601
-from telegram import Bot, Message, ReplyMarkup
+from telegram import Bot, Message
+from telegram._utils.types import ReplyMarkup
 
 from keel_telegram_bot.config import Config
 
@@ -14,6 +16,20 @@ LOGGER = logging.getLogger(__name__)
 CONFIG = Config()
 
 
+def _is_filtered_for(filters: List[Dict], chat_id: str, identifier: str) -> bool:
+    for config in filters:
+        filter_chat_id = config["chat_id"]
+        identifier_regex = config["identifier"]
+
+        if str(filter_chat_id) == str(chat_id):
+            identifier_pattern = re.compile(identifier_regex)
+            result = identifier_pattern.search(identifier)
+            if result is None:
+                return True
+
+    return False
+
+
 def flatten(data: List[List[Any]]) -> List[Any]:
     """
     Flattens a list of lists
@@ -33,8 +49,8 @@ def format_for_single_line_log(text: str) -> str:
     return " ".join(text.split())
 
 
-def send_message(bot: Bot, chat_id: str, message: str, parse_mode: str = None, reply_to: int = None,
-                 menu: ReplyMarkup = None) -> Message:
+async def send_message(bot: Bot, chat_id: str, message: str, parse_mode: str = None, reply_to: int = None,
+                       menu: ReplyMarkup = None) -> Message:
     """
     Sends a text message to the given chat
     :param bot: the bot
@@ -46,9 +62,12 @@ def send_message(bot: Bot, chat_id: str, message: str, parse_mode: str = None, r
     """
     from emoji import emojize
 
-    emojized_text = emojize(message, use_aliases=True)
-    return bot.send_message(chat_id=chat_id, parse_mode=parse_mode, text=emojized_text, reply_to_message_id=reply_to,
-                            reply_markup=menu)
+    emojized_text = emojize(message)
+    return await bot.send_message(
+        chat_id=chat_id, parse_mode=parse_mode, text=emojized_text,
+        reply_to_message_id=reply_to,
+        reply_markup=menu
+    )
 
 
 def fuzzy_match(term: str, choices: List[Any], limit: int = None, key=lambda x: x, ignorecase: bool = True) -> List[
@@ -110,7 +129,8 @@ def approval_to_str(data: dict) -> str:
     now_utc = datetime.now().replace(microsecond=0).astimezone(tz=timezone.utc)
     deadline_diff = timedelta(seconds=(deadline.replace(microsecond=0) - now_utc).total_seconds())
 
-    deadline_str = deadline_diff_to_str(deadline_diff)
+    deadline_abs_str = deadline.strftime('%m/%d %H:%M:%S')
+    deadline_remaining_str = deadline_diff_to_str(deadline_diff)
 
     text = "\n".join([
         f"<b>{message}</b>",
@@ -118,7 +138,7 @@ def approval_to_str(data: dict) -> str:
         f"Identifier: {identifier}",
         f"Version: {current_version} -> {new_version}",
         f"Votes: {votes_received}/{votes_required}",
-        f"Deadline in: {deadline} ({deadline_str})"
+        f"Expires: {deadline_abs_str} ({deadline_remaining_str})"
     ])
 
     return text
@@ -143,4 +163,4 @@ def deadline_diff_to_str(deadline_diff) -> str:
     if seconds:
         units.append(f"{seconds}s")
 
-    return " ".join(units)
+    return "".join(units)
